
#include <stdlib.h>
#include <memory.h>
#include <stdio.h>

#include "particle.h"

#define FALSE 0
#define TRUE 1

extern void drawflare(float px, float py, float pz, float size, float red, float green, float blue);
extern void drawdirt(float px, float py, float pz);

particle_t* particle;
long particles;

void particles_clear()
{
	if (particle)
		memset(particle, 0, particles * sizeof(particle_t));
}

int particles_alloc(long number)
{
	if (particle)
		free(particle);
	if (particle = malloc(number * sizeof(particle_t)))
	{
		particles = number;
		particles_clear();
		return TRUE;
	}
	return FALSE;
}

//long lastparticle; // used by spawning code

void particle_spawn(int rendertype, float starttime, float life, float size, float x, float y, float z, float vx, float vy, float vz, float color_r, float color_g, float color_b, float color_a, float decaystart, float decayend, float weightstart, float weightend, float pressurestart, float pressureend, float mass)
{
	long i;
/*
	if (lastparticle >= particles)
		lastparticle = 0;
	for (i = lastparticle + 1;i != lastparticle;i++)
	{
		if (i >= particles) // loop around
			i = 0;
*/
	for (i = 0;i < particles;i++)
	{
		if (!particle[i].active)
		{
			particle[i].active = TRUE;
			particle[i].rendertype = rendertype;
			particle[i].time = starttime;
			particle[i].life = particle[i].lifetime = life;
			particle[i].size = size;
			particle[i].x = x;
			particle[i].y = y;
			particle[i].z = z;
			particle[i].vx = vx;
			particle[i].vy = vy;
			particle[i].vz = vz;
			particle[i].color_red = color_r;
			particle[i].color_green = color_g;
			particle[i].color_blue = color_b;
			particle[i].color_alpha = color_a;
			particle[i].decaystart = decaystart;
			particle[i].decayend = decayend;
			particle[i].starttime = starttime;
			particle[i].weightstart = weightstart;
			particle[i].weightend = weightend;
			particle[i].pressurestart = pressurestart;
			particle[i].pressureend = pressureend;
			particle[i].mass = mass;
			return;
		}
	}
	printf("particle_spawn: ran out of particles\n");
}

extern float gravity_x, gravity_y, gravity_z;

void particles_advance(float currenttime)
{
	long i;
	float time;
	// kill off dead particles and update life values
	for (i = 0;i < particles;i++)
	{
		if (particle[i].active)
		{
			if ((particle[i].lifetime + particle[i].starttime) > currenttime)
			{
				particle[i].lifeprogress = (currenttime - particle[i].starttime) / particle[i].lifetime;
				particle[i].life = particle[i].lifetime + particle[i].starttime - currenttime;
			}
			else
				particle[i].active = FALSE;
		}
	}

	// apply pressure physics
	for (i = 0;i < particles;i++)
	{
		if (particle[i].active)
		{
			float pressure, weight;
			pressure = (particle[i].pressureend - particle[i].pressurestart) * particle[i].lifeprogress + particle[i].pressurestart;
			weight = (particle[i].weightend - particle[i].weightstart) * particle[i].lifeprogress + particle[i].weightstart;
			if (pressure < 0.0f)
				pressure = 0.0f;
			if (weight < 0.0f)
				weight = 0.0f;
			pressure -= weight;
			if (fabs(pressure) >= 0.01)
			{
				int j;
				// turn pressure into acceleration
				pressure *= currenttime - particle[i].time;
				for (j = 0;j < particles;j++)
				{
					if (particle[j].active && j != i && particle[j].mass > 0.0f)
					{
						float distx, disty, distz, distance2, scale;
						distx = particle[j].x - particle[i].x;
						disty = particle[j].y - particle[i].y;
						distz = particle[j].z - particle[i].z;
						distance2 = distx*distx+disty*disty+distz*distz;
						scale = particle[j].mass * pressure / (distance2+1);
						particle[j].vx += distx * scale;
						particle[j].vy += disty * scale;
						particle[j].vz += distz * scale;
					}
				}
			}
		}
	}

	// physics
	for (i = 0;i < particles;i++)
	{
		if (particle[i].active)
		{
			time = currenttime - particle[i].time;
			particle[i].time = currenttime;
			// a minimum of 100fps gravity processing (coarse calculations look bad)
			if (gravity_x || gravity_y || gravity_z)
			{
				while (time >= 0.01f)
				{
					particle[i].vx -= gravity_x * 0.01f;
					particle[i].x += particle[i].vx * 0.01f;
					particle[i].vy -= gravity_y * 0.01f;
					particle[i].y += particle[i].vy * 0.01f;
					particle[i].vz -= gravity_z * 0.01f;
					particle[i].z += particle[i].vz * 0.01f;
					time -= 0.01f;
				}
				// do the remaining time
				particle[i].vx -= gravity_x * time;
				particle[i].x += particle[i].vx * time;
				particle[i].vy -= gravity_y * time;
				particle[i].y += particle[i].vy * time;
				particle[i].vz -= gravity_z * time;
				particle[i].z += particle[i].vz * time;
			}
			else
			{
				particle[i].x += particle[i].vx * time;
				particle[i].y += particle[i].vy * time;
				particle[i].z += particle[i].vz * time;
			}
		}
	}
}
